from typing import Union, Any, List, Set, Dict, Tuple, Optional
import json

from prosto.utils import *

from prosto.Prosto import *
from prosto.Table import *
from prosto.Column import *
from prosto.Data import *

import logging
log = logging.getLogger('prosto')


class Table:
    """The class represents one table. Table data is stored in columns."""

    table_no = 0

    def __init__(self, prosto, definition):
        """
        Create a new table object using its definition.

        :param prosto: Prosto context object this table belongs to
        :param definition: Table definition as a dict
        """

        # Assign id
        self.id = definition.get('id', None)
        if self.id is None:
            self.id = "___table___" + str(Table.table_no)
            definition['id'] = self.id
            Table.table_no += 1

        self.prosto = prosto
        self.definition = definition

        # Here we store the real (physical) data for this table (all its attributes and columns)
        self.data = Data(self)

    def __repr__(self):
        return '['+self.id+']'

    def get_data(self) -> pd.DataFrame:
        return self.data.get_df()

    def get_column_data(self, column_name) -> pd.Series:
        return self.get_data()[column_name]

    #
    # Column getters
    #

    def get_column(self, column_name) -> Column:
        """Find a column definition object with the specified name"""
        return self.prosto.get_column(self.id, column_name)

    def get_columns(self) -> List[Column]:
        """Get a list of all columns of this table."""
        return self.prosto.get_columns(self.id)

    def evaluate(self) -> None:
        """Find a table operation which generates this table and execute it."""
        tab_ops = self.prosto.get_table_operations(self.id)
        tab_ops[0].evaluate()

if __name__ == "__main__":
    pass



from typing import Union, Any, List, Set, Dict, Tuple, Optional
import json

from prosto.utils import *

from prosto.Prosto import *
from prosto.Table import *
from prosto.Column import *

import logging
log = logging.getLogger('prosto')


class Operation:
    """The class represents one operation."""

    operation_no = 0

    def __init__(self, prosto, definition):
        """
        Create a new operation object using its definition.

        :param prosto: Prosto context object this operation belongs to
        :param definition: Operation definition as a dict
        """

        # Assign id
        self.id = definition.get('id', None)
        if self.id is None:
            self.id = "___operation___" + str(Operation.operation_no)
            definition['id'] = self.id
            Operation.operation_no += 1

        self.prosto = prosto
        self.definition = definition
        self.operation = definition.get("operation")

    def __repr__(self):
        return '['+self.id+']'

    def get_columns(self) -> List[str]:
        """Get a list of input column names specified in this definition."""
        definition = self.definition
        inputs = definition.get('columns', [])

        # If a single name is provided (not a list), then we wrap it into a list
        if isinstance(inputs, str):
            inputs = [inputs]

        return inputs

    def get_outputs(self) -> List[str]:
        """Get a list of output column names generated by this definition."""
        definition = self.definition
        outputs = definition.get('outputs', [])

        # If a single name is provided (not a list), then we wrap it into a list
        if isinstance(outputs, str):
            outputs = [outputs]

        if not outputs:  # If outputs are not specified then use id
            outputs.append(definition.get('id'))

        return outputs


if __name__ == "__main__":
    pass



from typing import Union, Any, List, Set, Dict, Tuple, Optional

from prosto.Prosto import *
from prosto.Table import *
from prosto.Column import *
from prosto.TableOperation import *
from prosto.ColumnOperation import *
from prosto.Topology import *

import logging
log = logging.getLogger('prosto')


class Prosto:
    """The class represents a context which stores lists of tables, columns and operations."""

    prosto_no = 0

    def __init__(self, id):

        self.id = id
        if self.id is None:
            self.id = "___prosto___" + str(self.prosto_no)
            self.prosto_no += 1

        self.tables = []
        self.columns = []
        self.operations = []

        self.topology = None
        self.incremental = False

    def __repr__(self):
        return '['+self.id+']'

    #
    # Table methods
    #

    def create_table(self, table_name, attributes) -> Table:
        """Create a new table with no operation that populates it. The table is supposed to be populated using API."""

        # Create a table definition
        table_def = {
            "id": table_name,
            "attributes": attributes,
        }
        table = Table(self, table_def)
        self.tables.append(table)

        return table

    def get_table(self, table_name) -> Table:
        """Find a table with the specified name"""
        if not table_name: return None
        return next((x for x in self.tables if x.id == table_name), None)

    def get_tables(self, table_names) -> List[Table]:
        """Get a list of tables with the specified names"""
        if not table_names: return []
        if isinstance(table_names, str):
            table_names = [table_names]
        tables = filter(lambda x: x.id in table_names, self.tables)
        return list(tables)

    #
    # Column getters
    #

    def get_column(self, table_name, column_name) -> Column:
        """Find a column the specified name"""
        if not table_name: return None
        if not column_name: return None
        return next((x for x in self.columns if x.id == column_name and x.table.id == table_name), None)

    def get_columns(self, table_name, column_names=None) -> List[Column]:
        """Get a list of columns with the specified names. All columns belong to one table."""
        if not table_name: return None
        if not column_names:
            all_columns = filter(lambda x: x.table.id == table_name, self.columns)
            return all_columns
        if isinstance(column_names, str):
            column_names = [column_names]
        columns = filter(lambda x: x.id in column_names and x.table.id == table_name, self.columns)
        return list(columns)

    #
    # Operations
    #

    def get_table_operations(self, table_name) -> List[TableOperation]:
        """Find operations which generate the specified table. Such operations have this table name in its outputs."""
        return [x for x in self.operations if isinstance(x, TableOperation) and table_name in x.definition.get("outputs", [])]

    def get_column_operations(self, table_name, column_name) -> List[ColumnOperation]:
        """Find operations which generate the specified column. Such operations have this column name in its outputs as well as the specified table name (each column operation has a table field)."""
        return [x for x in self.operations if isinstance(x, ColumnOperation) and column_name in x.definition.get("outputs", []) and table_name == x.definition.get("table")]

    #
    # Table operations
    #

    def populate(
            self,
            table_name, attributes,
            func, tables=None, model=None, input_length='table'
    ) -> Table:
        """Create a new populate table."""

        # Create a table definition
        table_def = {
            "id": table_name,
            "attributes": attributes,
        }
        table = Table(self, table_def)
        self.tables.append(table)

        # Create operation definition
        operation_def = {
            "id": None,
            "operation": 'populate',

            "outputs": [table_name],

            "function": func,
            "tables": tables,
            "model": model,
            "input_length": input_length,
        }
        operation = TableOperation(self, operation_def)
        self.operations.append(operation)

        return table

    def product(
            self,
            table_name, attributes,
            tables
    ) -> Table:
        """Create a new product table."""

        # Create a table definition
        table_def = {
            "id": table_name,
            "attributes": attributes,
        }
        table = Table(self, table_def)
        self.tables.append(table)

        # Create operation definition
        operation_def = {
            "id": None,
            "operation": 'product',

            "outputs": [table_name],

            "tables": tables,
        }
        operation = TableOperation(self, operation_def)
        self.operations.append(operation)

        return table

    def filter(
            self,
            table_name, attributes,
            func, tables, columns=None
    ) -> Table:
        """Create a new filter table."""

        # Create a table definition
        table_def = {
            "id": table_name,
            "attributes": attributes,
        }
        table = Table(self, table_def)
        self.tables.append(table)

        # Create operation definition
        operation_def = {
            "id": None,
            "operation": 'filter',

            "outputs": [table_name],

            "function": func,
            "tables": tables,
            "columns": columns,
        }
        operation = TableOperation(self, operation_def)
        self.operations.append(operation)

        return table

    def project(
            self,
            table_name, attributes,
            link, tables
    ) -> Table:
        """Create a new project table."""

        # Create a table definition
        table_def = {
            "id": table_name,
            "attributes": attributes,
        }
        table = Table(self, table_def)
        self.tables.append(table)

        # Create operation definition
        operation_def = {
            "id": None,
            "operation": 'project',

            "outputs": [table_name],

            "tables": tables,
            "link": link,
        }
        operation = TableOperation(self, operation_def)
        self.operations.append(operation)

        return table

    #
    # Column operations
    #

    def calculate(
            self,
            name, table,
            func, columns=None, model=None, input_length='column'
    ) -> Column:
        """Create a new calculate column."""

        # Create a column definition
        definition = {
            "id": name,
            "table": table,
        }
        column = Column(self, definition)
        self.columns.append(column)

        # Create operation definition
        operation_def = {
            "id": None,
            "operation": 'calculate',

            "table": table,
            "outputs": [name],

            "function": func,
            "columns": columns,
            "model": model,
            "input_length": input_length,
        }
        operation = ColumnOperation(self, operation_def)
        self.operations.append(operation)

        return column

    def link(
            self,
            name, table, type,
            columns, linked_columns=None
    ) -> Column:
        """Create a new link column."""

        # Create a column definition
        definition = {
            "id": name,
            "table": table,
            "type": type,
        }
        column = Column(self, definition)
        self.columns.append(column)

        # Create operation definition
        operation_def = {
            "id": None,
            "operation": 'link',

            "table": table,
            "outputs": [name],

            "columns": columns,
            "linked_columns": linked_columns,
        }
        operation = ColumnOperation(self, operation_def)
        self.operations.append(operation)

        return column

    def merge(
            self,
            name, table,
            columns
    ) -> Column:
        """
        Create a new merge column.

        A merge column materializes a column in some other table which is accessed via a column path.
        The column path is a sequence of link columns (except for maybe the last one).
        The first segments starts from this table and the last segment represents a column to be merged.
        """

        # Create a column definition
        definition = {
            "id": name,
            "table": table,
        }
        column = Column(self, definition)
        self.columns.append(column)

        # Create operation definition
        operation_def = {
            "id": None,
            "operation": 'merge',

            "table": table,
            "outputs": [name],

            "columns": columns,
        }
        operation = ColumnOperation(self, operation_def)
        self.operations.append(operation)

        return column

    def roll(
            self,
            name, table,
            window,
            func, columns=None, model=None, input_length='column'
    ) -> Column:
        """Create a new rolling aggregation column."""

        # Create a column definition
        definition = {
            "id": name,
            "table": table,
        }
        column = Column(self, definition)
        self.columns.append(column)

        # Create operation definition
        operation_def = {
            "id": None,
            "operation": 'rolling',

            "table": table,
            "outputs": [name],

            # How to group
            "window": window,

            # How to aggregate
            "function": func,
            "columns": columns,
            "model": model,
            "input_length": input_length,
        }
        operation = ColumnOperation(self, operation_def)
        self.operations.append(operation)

        return column

    def aggregate(
            self,
            name, table,
            tables, link,
            func, columns=None, model=None, input_length='column'
    ) -> Column:
        """Create a new aggregate column."""

        # Create a column definition
        definition = {
            "id": name,
            "table": table,
        }
        column = Column(self, definition)
        self.columns.append(column)

        # Create operation definition
        operation_def = {
            "id": None,
            "operation": 'aggregate',

            "table": table,
            "outputs": [name],

            # How to group
            "tables": tables,
            "link": link,

            # How to aggregate
            "function": func,
            "columns": columns,
            "model": model,
            "input_length": input_length,

            "initial_value": 0.0, # Pre-process like initial value
            "fillna_value": 0.0,  # Postprocess
        }
        operation = ColumnOperation(self, operation_def)
        self.operations.append(operation)

        return column

    #
    # Execution
    #

    def translate(self) -> Topology:
        self.topology = Topology(self)
        self.topology.translate()
        return self.topology

    def run(self) -> None:
        """Execute the whole workflow."""
        log.info("Start executing workflow '{}'.".format(self.id))

        # Translate if necessary
        if self.topology is None:
            self.translate()

        # Execute operations in the graph
        for layer in self.topology.layers:
            # Execute operations in one layer
            for op in layer:
                if isinstance(op, TableOperation):
                   op.evaluate()
                elif isinstance(op, ColumnOperation):
                    op.evaluate()
                else:
                    log.warning("Unknown element '{}' in the topology '{}'.".format(op.id, self.id))

        # Clear change status of all elements
        for tbl in self.tables:
            tbl.data.clear_change_status()
            tbl.data.gc()

        log.info("Finished executing workflow '{}'.".format(self.id))


if __name__ == "__main__":
    pass



import os
import sys
import argparse

import prosto as pr
#from prosto.Prosto import *

import logging
log = logging.getLogger('prosto')


def run(script_file):

    # TODO: Not implemented
    log.error("Script execution is not implemented.".format())

    return 0

def main(args = None):
    if not args: args = sys.argv[1:]

    programVersion = 'Version ' + pr.__version__
    programDescription = 'Prosto data processing toolkit' + programVersion

    parser = argparse.ArgumentParser(description=programDescription)
    parser.add_argument('-v', '--version', action='version', version=programVersion)

    parser.add_argument('-l', '--log', dest="loglevel", required=False, choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'], default='INFO', help="Set the logging level (default INFO)")

    parser.add_argument('script_file', type=str, help='Script file name')

    arguments = parser.parse_args(args)

    # Configure logging
    logging.basicConfig(stream=sys.stderr, level=arguments.loglevel, format='%(asctime)s - %(name)s - %(levelname)s: %(message)s')

    # Environment
    log.info(programDescription)

    exitcode = 1
    try:
        exitcode = run(arguments.script_file)
    except Exception as e:
        log.error("Error executing script file {}.".format(arguments.script_file))
        log.exception(e)

    logging.shutdown()

    return exitcode

if __name__ == "__main__":

    exitcode = main(sys.argv[1:])

    if(not exitcode):
        exit()
    else:
        exit(exitcode)



from typing import Union, Any, List, Set, Dict, Tuple, Optional
import json

from prosto.utils import *

from prosto.Prosto import *
from prosto.Table import *
from prosto.Column import *
from prosto.Operation import *

import logging
log = logging.getLogger('prosto')


class ColumnOperation(Operation):
    """The class represents one column operation."""

    def __init__(self, prosto, definition):
        super(ColumnOperation, self).__init__(prosto, definition)

    def get_dependencies(self) -> List[Union[Table, Column]]:
        """Get tables and columns this column depends upon."""
        definition = self.definition
        operation = definition.get('operation', 'UNKNOWN')

        output_table_name = definition.get('table')
        output_table = self.prosto.get_table(output_table_name)

        outputs = self.get_outputs()
        output_column_name = outputs[0]
        output_column = self.prosto.get_column(output_table_name, output_column_name)

        columns = self.get_columns()
        input_columns = self.prosto.get_columns(output_table_name, columns)

        dependencies = []

        # All derived columns depend on their table
        dependencies.append(output_table)

        if operation.lower().startswith('calc'):
            # Input column objects for which we need to find definitions
            dependencies.extend(input_columns)

        elif operation.lower().startswith('link'):
            # Target (linked) table has to be populated
            linked_table_name = output_column.definition.get('type', '')
            linked_table = self.prosto.get_table(linked_table_name)
            dependencies.append(linked_table)

            # Input (fact table) columns or column paths have to be evaluated
            dependencies.extend(input_columns)

            # Target columns have to be evaluated in order to contain values. However, they are supposed to be attributes and hence they will be set during population.
            linked_columns = definition.get('linked_columns', [])
            dependencies.extend(self.prosto.get_columns(linked_table_name, linked_columns))

        elif operation.lower().startswith('merg'):
            # Link column (first segment) has to be evaluated
            link_column_name = next(iter(columns), None)
            link_column = self.prosto.get_column(output_table_name, link_column_name)
            dependencies.append(link_column)

            # Linked column path (tail) in the linked table has to exist (recursion)
            linked_table_name = link_column.definition.get('type')
            linked_table = self.prosto.get_table(linked_table_name)
            linked_column_name = columns[1]

            linked_column = self.prosto.get_column(linked_table_name, linked_column_name)
            if linked_column:  # A linked column might not have a definition, e.g., an attribute
                dependencies.append(linked_column)
            # Here we assume that the tail dependencies will be retrieved separately.
            # Alternatively, we could retrieve them here using recursion

            # Lined table has to be populated. (Yet, it will be added to dependency by the link column.)
            dependencies.append(linked_table)

        elif operation.lower().startswith('roll'):
            # Columns to be aggregated
            dependencies.extend(input_columns)

        elif operation.lower().startswith('aggr'):
            # The fact table has to be already populated
            tables = definition.get('tables')
            source_table_name = tables[0]
            source_table = self.prosto.get_table(source_table_name)
            dependencies.append(source_table)

            # Group column
            link_column_name = definition.get('link')
            link_column = source_table.get_column(link_column_name)
            dependencies.append(link_column)

            # Measure columns
            dependencies.extend(self.prosto.get_columns(source_table_name, columns))

        else:
            return []

        return dependencies

    def evaluate(self) -> None:
        """
        Execute this column operation and evaluate the output column(s).

        A generic sequence of operations:
        - prepare the input slice by selecting input columns and input rows
        - convert the selected slice to the necessary data format expected by UDF
        - process input data by calling UDF or operation and returning some result
        - convert the result to our standard format
        - impose the result to our current data by overwriting the output columns and output values

        Notes:
        - There are two types of definitions: relying on UDFs (calc, roll, aggr), and not using UDFs (link, merge)
        - There are UDFs of two types: value or row based (returning a value or row), and column or table based (returning a whole column or table)
        """
        definition = self.definition
        operation = definition.get('operation', 'UNKNOWN')

        input_length = definition.get('input_length', 'UNKNOWN')  # value for value-based functions or column for column-based functions
        data_type = definition.get('data_type', 'Series')

        model = definition.get('model')

        output_table_name = definition.get('table')
        output_table = self.prosto.get_table(output_table_name)

        outputs = definition.get('outputs')
        output_column_name = outputs[0]
        output_column = self.prosto.get_column(output_table_name, output_column_name)

        data = output_table.get_data()

        log.info("---> Start evaluating '{}' column '{}'.".format(operation, self.id))

        #
        # Operations without UDF
        #

        # Link columns use their own definition format different from computational (functional) definitions
        if operation.lower().startswith('link'):
            out = self._evaluate_link()

            self._impose_output_columns(out)

            log.info("<--- Finish evaluating column '{}'".format(self.id))
            return

        # Compose columns use their own definition format different from computational (functional) definitions
        if operation.lower().startswith('merg'):
            out = self._evaluate_merge()

            self._impose_output_columns(out)

            log.info("<--- Finish evaluating column '{}'".format(self.id))
            return

        #
        # Operations with UDF
        #

        func_name = definition.get('function')
        if not func_name:
            log.error("Column function '{}' is not specified. Skip column definition.".format(func_name))
            return

        func = resolve_full_name(func_name)
        if not func:
            log.error("Cannot resolve user-defined function '{}'. Skip column definition.".format(func_name))
            return

        if operation.lower().startswith('calc'):
            # Determine input columns
            columns = self.get_columns()
            columns = get_columns(columns, data)
            if columns is None:
                log.error("Error reading column list. Skip column definition.")
                return

            # Validation: check if all explicitly specified columns available
            if not all_columns_exist(columns, data):
                log.error("Not all input columns available. Skip column definition.".format())
                return

            # Slice input according to the change status
            if self.prosto.incremental:
                data = output_table.data.get_added_slice(columns)
                range = output_table.data.added_range
            else:
                data = output_table.data.get_full_slice(columns)
                range = output_table.data.id_range()

            if input_length == 'value':
                out = self._evaluate_calc_value(func, data, data_type, model)
            elif input_length == 'column':
                out = self._evaluate_calc_column(func, data, data_type, model)
            else:
                log.error("Unknown input_type parameter '{}'.".format(input_length))
                return

        elif operation.lower().startswith('roll'):
            # Determine input columns
            columns = self.get_columns()
            columns = get_columns(columns, data)
            if columns is None:
                log.warning("Error reading input column list. Skip column definition.")
                return

            # Validation: check if all explicitly specified columns available
            if not all_columns_exist(columns, data):
                log.warning("Not all input columns available. Skip column definition.".format())
                return

            # Slice input according to the change status (incremental not implemented)
            data = output_table.data.get_full_slice(columns)
            range = output_table.data.id_range()

            if input_length == 'value':
                log.error("Accumulation is not implemented.".format())
                return
            elif input_length == 'column':
                out = self._evaluate_roll_column(func, data, data_type, model)
            else:
                log.error("Unknown input_type parameter '{}'.".format(input_length))
                return

        elif operation.lower().startswith('aggr'):
            #
            # Get parameters
            #
            tables = definition.get('tables')
            source_table_name = tables[0]
            source_table = self.prosto.get_table(source_table_name)
            if source_table is None:
                log.error("Cannot find the fact table '{}'.".format(source_table_name))
                return

            link_column_name = definition.get('link')
            link_column = source_table.get_column(link_column_name)
            if link_column is None:
                log.error("Cannot find the link column '{}'.".format(link_column_name))
                return

            data = source_table.get_data()  # Data (to be processed) is a (source) table which is different from the output table

            # Determine input columns
            columns = self.get_columns()
            columns = get_columns(columns, data)
            if columns is None:
                log.warning("Error reading input column list. Skip column definition.")
                return

            # Validation: check if all explicitly specified columns available
            if not all_columns_exist(columns, data):
                log.warning("Not all input columns available. Skip column definition.".format())
                return

            data = data[columns]  # Select only the specified *input* columns

            data_type = definition.get('data_type')

            # No incremental. Select full *output* range
            range = output_table.data.id_range()

            if input_length == 'value':
                log.error("Accumulation is not implemented.".format())
                return
            elif input_length == 'column':
                out = self._evaluate_group_column(func, data, data_type, model)
            else:
                log.error("Unknown input_type parameter '{}'.".format(input_length))
                return

        else:
            log.error("Unknown operation type '{}' in the definition of column '{}'.".format(operation, self.id))
            return

        #
        # Append the newly generated column(s) to this table
        #
        self._impose_output_columns(out, range)

        log.info("<--- Finish evaluating column '{}'".format(self.id))

    def _evaluate_calc_value(self, func, data, data_type, model):
        """Calculate column. Apply function to each row of the table."""

        #
        # Single input: Apply to a series. UDF will get single value
        #
        if isinstance(data, pd.Series) or ((isinstance(data, pd.DataFrame) and len(data.columns) == 1)):
            if isinstance(data, pd.DataFrame):
                ser = data[data.columns[0]]
            else:
                ser = data

            if model is None:
                out = pd.Series.apply(ser, func)  # Do not pass model to the function
            elif isinstance(model, (list, tuple)):
                out = pd.Series.apply(ser, func, *model)  # Model as positional arguments
            elif isinstance(model, dict):
                # Pass model by flattening dict (alternative: arbitrary Python object as positional or key argument). UDF has to declare the expected arguments
                out = pd.Series.apply(ser, func, **model)  # Model as keyword arguments
            else:
                out = pd.Series.apply(ser, func, args=(model,))  # Model as an arbitrary object

        #
        # Multiple inputs: Apply to a frame. UDF will get a row of values
        #
        elif isinstance(data, pd.DataFrame):
            # Notes:
            # - UDF expects one row as a data input (raw=True - ndarry, raw=False - Series)
            # - model (arguments) cannot be None, so we need to guarantee that we do not pass None

            if data_type == 'ndarray':
                out = pd.DataFrame.apply(data, func, axis=1, raw=True, **model)
            else:
                if model is None:
                    out = pd.DataFrame.apply(data, func, axis=1, raw=False)  # Do not pass model to the function
                elif isinstance(model, (list, tuple)):
                    out = pd.DataFrame.apply(data, func, axis=1, raw=False, *model)  # Model as positional arguments
                elif isinstance(model, dict):
                    out = pd.DataFrame.apply(data, func, axis=1, raw=False, **model)  # Model as keyword arguments
                else:
                    out = pd.DataFrame.apply(data, func, axis=1, raw=False, args=(model,))  # Model as an arbitrary object

        else:
            log.error("Unknown input data type '{}'".format(type(data_type).__name__))

        return out

    def _evaluate_calc_column(self, func, data, data_type, model):
        """Calculate column. Apply function to all inputs and return calculated column(s)."""

        #
        # Cast to the necessary argument type expected by the function
        #
        if data_type == 'ndarray':
            data_arg = data.values
            data_arg.reshape(-1, 1)
        elif (isinstance(data, pd.DataFrame) and len(data.columns) == 1):
            # data_arg = data
            data_arg = data[data.columns[0]]
        else:
            data_arg = data

        if isinstance(model, dict):
            out = func(data_arg, **model)  # Model as keyword arguments
        elif isinstance(model, (list, tuple)):
            out = func(data_arg, *model)  # Model as positional arguments
        else:
            out = func(data_arg, model)  # Model as an arbitrary object

        return out

    def _evaluate_link(self):
        """Link column. Output column will store ids (indexes) of the target table rows."""
        definition = self.definition

        #
        # Read and validate parameters
        #

        main_table_name = definition.get('table')
        main_table = self.prosto.get_table(main_table_name)

        outputs = definition.get('outputs')
        column_name = outputs[0]
        output_column = self.prosto.get_column(main_table_name, column_name)

        main_keys = self.get_columns()
        if not all_columns_exist(main_keys, main_table.get_data()):
            log.error("Not all key columns available in the link column definition.".format())
            return

        linked_table_name = output_column.definition.get('type', '')
        linked_table = self.prosto.get_table(linked_table_name)
        if not linked_table:
            log.error("Linked table '{}' cannot be found in the link column definition.".format(linked_table))
            return

        linked_columns = definition.get('linked_columns', [])
        if len(linked_columns) == 0:
            linked_columns = linked_table.definition.get("attributes", [])  # By default (e.g., for projection), we link to target table attributes
        if not all_columns_exist(linked_columns, linked_table.get_data()):
            log.error("Not all linked key columns available in the link column definition.".format())
            return

        #
        # 1. In the target (linked) table, convert its index into a normal column (because we can only merge normal columns and index)
        #
        index_column_name = '__row_id__' # It could be 'id', 'index' or whatever other convention
        linked_table.get_data()[index_column_name] = linked_table.get_data().index
        # df.reset_index(inplace=True).set_index('index', drop=False, inplace=True)  ä Alternative 1: reset will convert index to column, and then again create index
        # df = df.rename_axis('index1').reset_index() # Alternative 2: New index1 column will be created

        #
        # 2. Create left join on the specified keys
        #

        linked_prefix = column_name+'::'  # It will be prepended to each linked (secondary) column name

        out_df = pd.merge(
            main_table.get_data(),  # This table
            linked_table.get_data().rename(columns=lambda x: linked_prefix + x, inplace=False),  # Target table to link to. We rename columns (not in place - the original frame preserves column names)
            how='left',  # This (main) table is not changed - we attach target records
            left_on=main_keys,  # List of main table key columns
            right_on= [linked_prefix + x for x in linked_columns],  # List of target table key columns. Note that we renamed them above so we use modified names.
            left_index=False,
            right_index=False,
            #suffixes=('', linked_suffix),  # We do not use suffixes because they cannot be enforced (they are used only in the case of equal column names)
            sort=False  # Sorting decreases performance
        )

        # We do not need this column anymore - it was merged into the result as a link column
        linked_table.get_data().drop(columns=[index_column_name], inplace=True)

        #
        # 3. Rename according to our convention and store the result
        #

        # Rename our link column by using only specified column name
        out_df.rename({column_name+'::'+index_column_name: column_name}, axis='columns', inplace=True)

        out = out_df[column_name]

        return out

    def _evaluate_merge(self):
        """Merge column. Materialize a complex column path which is sequence of link columns ending with some target column."""
        definition = self.definition

        #
        # Read and validate parameters
        #
        output_table_name = definition.get('table')
        output_table = self.prosto.get_table(output_table_name)
        output_table_data = output_table.get_data()

        outputs = definition.get('outputs')
        output_column_name = outputs[0]
        output_column = self.prosto.get_column(output_table_name, output_column_name)

        column_segment_separator = '::'

        columns = self.get_columns()
        link_column_path = ''  # Column path composed of several separated column segment names
        df = output_table_data
        main_table = output_table
        for i, link_column_name in enumerate(columns):
            #
            # Build link column path by appending a new column segment name
            # Note that this path is a name of the column in the data frame and it does not have a definition
            #
            if i == 0:
                link_column_path = link_column_name
            else:
                link_column_path += column_segment_separator + link_column_name

            #
            # Stop condition if there is no further segment to merge.
            #
            if i == len(columns) - 1:
                # Rename the last merged column path to the desired column name as specified in the definition
                df.rename(columns={link_column_path: output_column_name}, inplace=True)
                out = df[output_column_name]
                break

            #
            # Find link column definition
            #
            link_column = main_table.get_column(link_column_name)

            #
            # Find the linked table
            #
            linked_table_name = link_column.definition.get('type')
            linked_table = self.prosto.get_table(linked_table_name)
            linked_table_data = linked_table.get_data()

            #
            # Find the target linked column in the linked table
            #
            linked_column_name = columns[i+1]
            #linked_column = linked_table.get_column(linked_column_name)

            #
            # Do merge
            #
            df = self._merge_two_columns(
                df,
                link_column_path,
                linked_table_data,
                linked_column_name
            )

            # Iterate
            main_table = linked_table

        return out

    def _merge_two_columns(self, source_df, link_column_name, target_df, linked_column_name):
        """Given source and target tables with link and linked columns, return a new column with the source table index and target column values."""

        linked_prefix = link_column_name + '::'  # It will prepended to each linked (secondary) column name

        out_df = pd.merge(
            source_df[[link_column_name]],
            target_df[[linked_column_name]].rename(columns=lambda x: linked_prefix + x, inplace=False),  # We rename target columns to avoid conflicts (not in place - the original frame preserves column names)
            how='left',  # This (source) table is not changed - we want to attach target records
            left_on=link_column_name,  # Its values are row ids stored in the linked table index
            right_on=None,
            left_index=False,
            right_index=True,  # Use index because link column stores index values
            #suffixes=('', linked_suffix),  # We do not use suffixes because they cannot be enforced (they are used only in the case of equal column names)
            sort=False  # Sorting decreases performance
        )
        # Here we get linked column names like 'Prefix::OriginalName'

        return out_df

    def _evaluate_roll_column(self, func, data, data_type, model):
        """Roll column. Apply aggregate function to each window defined on this same table for every record."""
        definition = self.definition

        #
        # Determine window size. The window parameter can be string, number or object (many arguments for rolling object)
        #
        window = definition.get('window')
        window_size = int(window)
        rolling_args = {'window': window_size}
        # TODO: try/catch with log message if cannot get window size

        #
        # Single input. Moving aggregation of one input column. Function will get a sub-series as a data argument
        #
        if len(data.columns) == 1:

            in_column = data.columns.to_list()[0]

            # Create a rolling object with windowing (row-based windowing independent of the number of columns)
            by_window = pd.DataFrame.rolling(data, **rolling_args)  # as_index=False - aggregations will produce flat DataFrame instead of Series with index

            # Apply function to all windows
            if data_type == 'ndarray':
                out = by_window[in_column].apply(func, raw=True, **model)
            else:
                out = by_window[in_column].apply(func, raw=False, **model)

        #
        # Multiple inputs. Function will get a sub-dataframe as a data argument
        #
        else:

            #
            # Workaround: create a new temporary data frame with all row ids, create a rolling object by using it, apply UDF to it, the UDF will get a window/group of row ids which can be then used to access this window rows from the main data frame:
            # Link: https://stackoverflow.com/questions/45928761/rolling-pca-on-pandas-dataframe
            #

            df_idx = pd.DataFrame(np.arange(data.shape[0]))  # Temporary data frame with all row ids like 0,1,2,...
            idx_window = df_idx.rolling(**rolling_args)  # Create rolling object from ids-frame

            # Auxiliary function creates a subframe with data and passes it to the user function
            def window_fn(ids, user_f):
                return user_f(data.iloc[ids])

            out = idx_window.apply(lambda x: window_fn(x, func), raw=False)  # Both Series and ndarray work (for iloc)

        return out

    def _evaluate_group_column(self, func, data, data_type, model):
        """Group column. Apply aggregate function to each group of records of the fact table."""
        definition = self.definition

        #
        # Group by the values (ids) of the link column. All facts with the same id in the link column belong to one group
        #
        gb = self._get_or_create_groupby()

        if len(data.columns) == 0:  # Special case: no input columns (or function is size()
            out = gb.size()
        if len(data.columns) == 1:  # Single input: udf will get a sub-series with fact values
            out = gb[data.columns[0]].agg(func, **model)  # Apply function to all groups
        else:  # Multiple inputs. Function will get a sub-dataframe as a data argument
            # TODO:
            pass

        return out

    def _get_or_create_groupby(self):
        """Each link column stores a pandas groupby object. Return or build such a groupby object for the (already evaluated) link column specified in this definition."""

        definition = self.definition

        tables = definition.get('tables')
        source_table_name = tables[0]
        source_table = self.prosto.get_table(source_table_name)

        link_column_name = definition.get('link')
        link_column = self.prosto.get_column(source_table_name, link_column_name)

        if link_column.groupby is not None:
            return link_column.groupby

        # Use link column (with target row ids) to build a groupby object (it will build a group for each target row id)
        try:
            gb = source_table.get_data().groupby(link_column_name, as_index=True)
            # Alternatively, we could use target keys or main keys
        except Exception as e:
            log.error("Error grouping input table using the specified column(s).".format())
            log.debug(e)
            raise e

        # TODO: We might want to remove a group for null value (if it is created by the groupby constructor)

        link_column.groupby = gb

        return gb

    def _impose_output_columns(self, out, range=None):
        """
        Append the specified column(s) to the data frame of the output table.
        This function will impose the input data frame onto the existing data.
        It will overwrite existing values with the same ids and columns as in input data frame
        Other values in this table which are not present in the input data frame will be not changed.
        None range means full id range.
        """
        definition = self.definition

        outputs = self.get_outputs()
        output_table_name = definition.get('table')
        output_table = self.prosto.get_table(output_table_name)

        fillna_value = definition.get('fillna_value')

        #
        # Change format to dataframe
        #
        if isinstance(out, pd.DataFrame):
            pass
        elif isinstance(out, pd.Series):
            out = pd.DataFrame(out)  # Series name will be column name
        elif isinstance(out, (list,tuple)):
            # TODO: Convert a list of series or data frames into one data frame. They all have to have same index.
            log.error("List (of series) as a result of evaluation is currently not supported.".format())
            return
        elif isinstance(out, np.ndarray):
            # TODO: Convert ndarray into dataframe. Main problem is that we need to know the index and then assume that the values in ndarray are sequential.
            log.error("NumPy array as a result of evaluation is currently not supported.".format())
            return

        #
        # Assign (custom) column names
        #
        if len(out.columns) > len(outputs):
            log.error("Operation returned {} columns, which is more than specified in its definition for its output.".format(len(out.columns)))
            return

        if len(out.columns) < len(outputs):
            log.warning("Operation returned {} columns, which is less than specified in its definition for its output.".format(len(out.columns)))

        out.columns = outputs[0:len(out.columns)]

        #
        # Write the result to the data by overwriting cells
        #
        output_table.data.set_column_values_for_range(out, range, fillna_value)


if __name__ == "__main__":
    pass



from typing import Union, Any, List, Set, Dict, Tuple, Optional
import os
import sys
import types
import inspect
import importlib
import importlib.util
import functools

import pandas as pd
import numpy as np

import logging
log = logging.getLogger('prosto')


#
# Columns and frames
#

def get_columns(names, df=None) -> List[str]:
    """Produce a list of column names by also validating them against the data frame."""
    result = []

    if isinstance(names, str):  # A single column name
        result.append(names)

    elif isinstance(names, int) and df is not None:  # A single column number
        result.append(df.columns[names])

    elif isinstance(names, (list, tuple)):  # A list of column names
        for col in names:
            if isinstance(col, str):
                result.append(col)
            elif isinstance(col, int) and df is not None:
                result.append(df.columns[col])
            else:
                log.error("Error reading column '{}'. Names have to be strings or integers.".format(str(col)))
                return None

        # Process default (auto) values
        if len(result) == 0 and df is not None:  # Explicit empty list = ALL columns
           result = get_all_columns(df)

    elif isinstance(names, dict):  # An object specifying which columns to select
        exclude = names.get("exclude")
        if not isinstance(exclude, dict):
            exclude_columns = get_columns(exclude, df)
        else:
            log.error(f"Error reading column '{exclude}'. Excluded columns have to be (a list of) strings or integers.")
            return None

        # Get all columns and exclude the specified ones
        all_columns = get_all_columns(df)
        result = [x for x in all_columns if x not in exclude_columns]

    else:
        log.error(f"Column names have to be a list of strings or a string.")
        return None

    #
    # Validate
    #

    # Check that all columns are available
    if df is None:
        return result
    elif isinstance(df, pd.DataFrame):
        out = []
        for col in result:
            if col in df.columns:
                out.append(col)
            else:
                log.warning(f"Column '{str(col)}' cannot be found. Skip column.")
        return out

    elif isinstance(df, pd.core.groupby.groupby.DataFrameGroupBy):
        out = []
        for col in result:
            col_exists = False
            try:
                df.__getattr__(col)
                col_exists = True
            except:
                pass

            if col_exists:
                out.append(col)
            else:
                log.warning(f"Column '{str(col)}' cannot be found. Skip column.")
        return out

    return result

def get_all_columns(df) -> List[str]:
    if df is None:
        return []
    elif isinstance(df, pd.DataFrame):
        return df.columns.tolist()
    elif isinstance(df, pd.core.groupby.groupby.DataFrameGroupBy):
        # TODO: We need to exclude key columns which are used for gropuing
        return df.obj.columns.tolist()
    else:
        return None

def all_columns_exist(names, df) -> bool:
    all_columns_available = True
    for col in names:
        if col not in df.columns:
            all_columns_available = False
            log.warning(f"Column '{col}' is not available.")
            break
    if not all_columns_available:
        return False
    return True

#
# Function resolution
#

def resolve_full_name(full_name):
    """
    Resolve the specified name or definition of the function to a reference.
    Fully qualified name consists of module name and function name separated by a colon, for example:  'mod1.mod2.mod3:class1.class2.func1.func2'.
    """

    if not full_name:
        return None

    elif isinstance(full_name, (types.FunctionType, types.BuiltinFunctionType, functools.partial)):
        return full_name

    elif full_name.strip().startswith('lambda '):
        try:
            func = eval(full_name)
        except Exception as e:
            log.error(f"Error translating lambda function: {full_name}.")
            log.debug(e)
            return None
        return func

    elif full_name.strip().startswith('def '):
        pass  # TODO: Standard function definition

    else:  # Function name
        mod_and_func = full_name.split(':', 1)
        mod_name = mod_and_func[0] if len(mod_and_func) > 1 else None
        func_name = mod_and_func[-1]

        if mod_name:
            mod = resolve_module(mod_name)
            if mod is None: return None
            func = resolve_name_in_mod(func_name, mod)
            return func
        else:
            pass # TODO: Module is not specified. Search in all modules

    return None

def all_modules():
    modules = []
    return modules

def resolve_module(mod_name: str):
    mod = sys.modules.get(mod_name)

    if mod:
        return mod

    try:
        mod = importlib.import_module(mod_name)
    except Exception as e:
        pass

    return mod

def resolve_name_in_mod(func_name: str, mod):
    # Split full name into segments (classes and functions)
    name_path = func_name.split('.')

    # Sequentially resolve each next segment in the result of the previous segment starting from the specified module
    last_segment = mod
    for i in range(len(name_path)):
        name_segment = name_path[i]
        ref_segment = None

        try:
            ref_segment = getattr(last_segment, name_segment)
            """
            for key, val in mod.__dict__.items():
                if not inspect.isclass(val): continue

                members = inspect.getmembers(val, predicate=inspect.ismethod)  # A list of all members of the class
                for n, m in members:
                    if n == func_name: return m
            """
        except AttributeError as e:
            pass

        if ref_segment is None:
            return None
        else:
            last_segment = ref_segment

    return last_segment

def import_modules(imports):
    modules = []

    for mod_name in imports:
        mod = None
        try:
            mod = importlib.import_module(mod_name)
        except ImportError as ie:
            pass

        if mod:
            modules.append(mod)
            continue  # Module found and imported

        # Try to import from source file
        # See: https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
        try:
            file_path = mod_name.replace('.', '/')
            file_path = file_path + '.py'
            spec = importlib.util.spec_from_file_location(mod_name, file_path)
            mod = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(mod)
        except ImportError as ie:
            pass

        if mod:
            modules.append(mod)
            sys.modules[mod_name] = mod
            continue

        log.warning(f"Cannot import module '{mod_name}'. Ignored. This can cause errors later if its functions are used in the workflow")

    return modules


if __name__ == "__main__":
    pass



import json
import copy

from prosto.utils import *

from prosto.Prosto import *
from prosto.Table import *
from prosto.Column import *
from prosto.TableOperation import *
from prosto.ColumnOperation import *
from prosto.Data import *

import logging
log = logging.getLogger('prosto')


class Topology:
    """Topology is a graph of operations built taking into account their dependencies."""

    def __init__(self, prosto):

        self.prosto = prosto  # If context can be modified then make a copy: copy.deepcopy(prosto)

        self.layers = []  # Graph of operations
        self.elem_layers = []  # Graph of elements

    def translate(self) -> None:
        """Build a graph of operations by analyzing table and column dependencies."""

        #
        # Collect all operations into one list
        #
        all_operations = [x for x in self.prosto.operations]

        #
        # Build graph of operations by analyzing dependencies
        #

        # Empty collection of already processed elements (they can be simultaneously removed from all)
        done = []

        # Topology to be built is a list of layers in the order of execution of their operations.
        # First layer does not have dependencies. Second layer depends on the operations in the first layer and so on.
        layers = []
        while True:
            layer = []  # List of operations which depend on only operations in some previous layers

            # We will build this new layer from the available (not added to previous layers) operations
            for op in all_operations:

                if op in done:
                    continue

                #
                # Find dependencies of this operation
                #
                if isinstance(op, (TableOperation, ColumnOperation)):
                    dep_elems = op.get_dependencies()  # Get all element definitions this element depends upon
                else:
                    log.error("Operation '{}' with unknown class found while building topology.".format(op.id))
                    dep_elems = []

                #
                # Find operations which generate these elements
                #
                dep_ops = []
                for dep in dep_elems:
                    if isinstance(dep, Table):
                        ops = self.prosto.get_table_operations(dep.id)
                    elif isinstance(dep, Column):
                        ops = self.prosto.get_column_operations(dep.table.id, dep.id)
                    else:
                        log.error("Element '{}' with unknown class found while building topology (only Table and Column are possible).".format(dep.id))
                        ops = []
                    dep_ops.extend(ops)

                #
                # If all dependencies have been processed (belong to previous layers) then add this element to the current layer
                #
                if set(dep_ops) <= set(done):
                    layer.append(op)

            if len(layer) == 0:
                break

            layers.append(layer)
            done.extend(layer)

        # Layers of operations
        self.layers = layers

        # Layers of tables/columns
        elem_layers = []
        for layer in layers:
            elem_layer = []
            for op in layer:
                # Find element (column or table) this operation produces
                outputs = op.get_outputs()
                if isinstance(op, TableOperation):  # Find table
                    tables = self.prosto.get_tables(outputs)
                    elem_layer.extend(tables)

                    # Allocate/initialize data and other resources
                    for tab in tables:
                        tab.data = Data(tab)

                elif isinstance(op, ColumnOperation):  # Find column
                    table_name = op.definition.get("table")
                    columns = self.prosto.get_columns(table_name, outputs)
                    elem_layer.extend(columns)
            elem_layers.append(elem_layer)

        self.elem_layers = elem_layers


if __name__ == '__main__':
    pass



import setuptools

# import ``__version__`` from code base
exec(open('prosto/__init__.py').read())

setuptools.setup(
    name='prosto',
    version=__version__,

    # descriptive metadata for upload to PyPI
    description='Data processing toolkit radically changing the way data is processed',
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    author='Prosto Data',
    author_email='52108119+prostodata@users.noreply.github.com',
    license='MIT License',
    keywords = ['data processing', 'analytics', 'data science', 'pandas', 'map-reduce', 'feature engineering', 'business intelligence'],
    url='https://github.com/prostodata/prosto',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers ",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
    ],

    test_suite='nose.collector',
    tests_require=['nose'],

    # dependencies
    install_requires=[
        'numpy',
        'pandas',
    ],
    zip_safe=True,

    # package content (what to include)
    packages=setuptools.find_packages(),
    #packages=setuptools.find_packages(exclude=("tests",)),
    #packages=["prosto", "notebooks", "tests"],

    package_data={
        # If any package contains *.txt or *.rst files, include them:
        '': ['*.md', '*.txt', '*.rst'],
        # And include any *.msg files found in the 'hello' package, too:
        #'hello': ['*.msg'],
    },

)



import json

from prosto.utils import *

from prosto.Prosto import *
from prosto.Table import *
from prosto.Column import *

import logging
log = logging.getLogger('prosto')


class Column:
    """The class represents one column definition."""

    column_no = 0

    def __init__(self, prosto, definition):
        """
        Create a new column object using its definition.

        :param prosto: Prosto context object this table belongs to
        :param definition: Column definition as a dict
        """

        # Assign id
        self.id = definition.get('id', None)
        if self.id is None:
            self.id = "___column___" + str(Column.column_no)
            definition['id'] = self.id
            Column.column_no += 1

        self.prosto = prosto
        self.definition = definition

        # Resolve from-table name
        table_name = self.definition.get("table", "")
        self.table = self.prosto.get_table(table_name)

        # Resolve to-table name
        table_name = self.definition.get("type", "")
        if table_name:  # It is link column (the type-table must already exist)
            self.type = self.prosto.get_table(table_name)
        else:  # Primitive type
            self.type = None

        # Link columns store here a groupby object which is then (re)used by aggregate columns
        self.groupby = None


    def __repr__(self):
        return '[' + self.table.id + '::' + self.id+']'

    def evaluate(self) -> None:
        """Find a column operation which generates this column and execute it."""
        col_ops = self.prosto.get_column_operations(self.table.id, self.id)
        col_ops[0].evaluate()


if __name__ == "__main__":
    pass



from typing import Union, Any, List, Set, Dict, Tuple, Optional
import json

from prosto.utils import *

from prosto.Prosto import *
from prosto.Table import *
from prosto.Column import *
from prosto.Operation import *

import logging
log = logging.getLogger('prosto')


class TableOperation(Operation):
    """The class represents one table operation."""

    def __init__(self, prosto, definition):
        super(TableOperation, self).__init__(prosto, definition)

    def get_dependencies(self) -> List[Union[Table, Column]]:
        """Get tables and columns this table depends upon."""
        definition = self.definition
        operation = definition.get('operation', 'UNKNOWN')

        outputs = self.get_outputs()
        output_table_name = outputs[0]
        output_table = self.prosto.get_table(output_table_name)

        tables = self.definition.get("tables")
        input_tables = self.prosto.get_tables(tables)

        dependencies = []

        dependencies.extend(input_tables)

        if operation.lower().startswith('popu'):
            # We want to evaluate *all* columns of the base table except for those which depend on this table
            for tab in input_tables:
                # List all derived columns of the base table
                for col in tab.get_columns():
                    if col.get("operation").startswith("attr"):
                        continue
                    # Get its dependencies
                    deps = col.get_dependencies()
                    if self.id in deps:
                        continue  # Do NOT include in dependencies because it depends on this table

                    dependencies.append(col)

        elif operation.lower().startswith('prod'):
            pass

        elif operation.lower().startswith('filt'):
            base_table = input_tables[0]

            # Base table filter column
            columns = definition.get('columns')
            filter_column_name = columns[0]
            filter_column = base_table.get_column(filter_column_name)
            dependencies.append(filter_column)

        elif operation.lower().startswith('proj'):
            # Source table
            source_table_name = tables[0]
            source_table = input_tables[0]

            # Source table link column input keys (not the link column itself)
            link_column_name = definition.get('link')
            link_column = source_table.get_column(link_column_name)

            # Input columns of the link column
            link_column_ops = self.prosto.get_column_operations(source_table_name, link_column_name)
            link_column_op = link_column_ops[0]
            source_keys = link_column_op.get_columns()
            dependencies.extend(self.prosto.get_columns(source_table_name, source_keys))

        else:
            log.warning("Unknown operation type '{}' in the definition of table '{}'.".format(operation, self.id))
            return

        return dependencies

    def evaluate(self) -> None:
        """Execute this operation by populating the output table. Only attribute columns are filled with values."""
        definition = self.definition
        operation = definition.get('operation', 'UNKNOWN')

        input_length = definition.get('input_length', 'UNKNOWN')  # value for row-based functions or table for column-based functions
        data_type = definition.get('data_type', 'DataFrame')

        model = definition.get('model')

        outputs = definition.get('outputs')
        output_table_name = outputs[0]
        output_table = self.prosto.get_table(output_table_name)

        log.info("===> Start populating '{}' table '{}'".format(operation, output_table.id))

        if operation.lower().startswith('noop'):
            new_data = None
        if operation.lower().startswith('popu'):

            if input_length == 'row':
                new_data = self._evaluate_populate_row()
            elif input_length == 'table':
                new_data = self._evaluate_populate_table()
            else:
                log.error("Unknown input_type parameter '{}'.".format(input_length))
                return

        elif operation.lower().startswith('prod'):
            new_data = self._evaluate_product()

        elif operation.lower().startswith('filt'):
            new_data = self._evaluate_filter()

        elif operation.lower().startswith('proj'):
            new_data = self._evaluate_project()

        else:
            log.warning("Unknown operation type '{}' in the definition of table '{}'.".format(operation, self.id))
            return

        if new_data is not None:
            output_table.data.remove_all()
            output_table.data.add(new_data)

        log.info("<=== Finish populating table '{}'".format(output_table.id))

    def _evaluate_populate_row(self):
        """The function is applied to one row (from an input table) and generates a sub-table which will be appnded to the result."""
        definition = self.definition
        log.error("Row-based table population not supported.".format())

    def _evaluate_populate_table(self):
        """The result dataframe is genreated by the specified function using dataframe from input tables and model parameters."""
        definition = self.definition

        #
        # Stage 1. Resolve the function
        #
        func_name = definition.get('function')
        if not func_name:
            log.error("Table function '{}' is not specified. Skip table definition.".format(func_name))
            return

        func = resolve_full_name(func_name)
        if not func:
            log.error("Cannot resolve user-defined function '{}'. Skip table definition.".format(func_name))
            return

        #
        # Stage 2. Prepare input data
        #
        tables = self.definition.get("tables")
        tables = self.prosto.get_tables(tables)
        if not tables: tables = []

        #
        # Stage 3. Prepare argument object to pass to the function as the second argument
        #
        model = definition.get('model')

        #
        # Stage 4. Apply function
        #
        out = None
        if len(tables) == 0:
            if model:
                out = func(**model)
            else:
                out = func()

        elif len(tables) == 1:  # Pass a single data frame
            if model:
                out = func(tables[0].get_data(), **model)
            else:
                out = func(tables[0].get_data())

        else:  # Pass a list of data frames
            if model:
                out = func([t.get_data() for t in tables], **model)
            else:
                out = func([t.get_data() for t in tables])

        #
        # Stage 5. Check that the result includes all declared attributes
        #
        attributes = definition.get("attributes", [])
        columns = out.columns.tolist()
        for att in attributes:
            if att not in columns:
                log.error("Declared attribute '{}' is not in the populated table data.".format(att))
                return

        #
        # Stage 6. Ensure that it has an integer index
        #
        out.reset_index(inplace=True, drop=True)

        return out

    def _evaluate_product(self):
        """The output table is a Cartesian product of the input tables with attributes pointing to the input records they are made of."""
        definition = self.definition

        outputs = definition.get('outputs')
        output_table = self.prosto.get_table(outputs[0])

        attributes = output_table.definition.get("attributes", [])

        #
        # Stage 1. Prepare input data
        #
        tables = self.definition.get("tables")
        if not tables:
            log.error("Table product operation must specify at least one table in the 'tables' field.".format())
            return
        if len(tables) != len(attributes):
            log.error("Number of input tables must be equal to the number of attributes in product table definition.".format())
            return

        tables = self.prosto.get_tables(tables)
        table_datas = [x.get_data() for x in tables]

        #
        # Compute Cartesian product of all tables on their indexes
        #

        table_indexes = [x.index for x in table_datas]

        index = pd.MultiIndex.from_product(iterables=table_indexes, names=attributes)
        out = pd.DataFrame(index=index)
        out.reset_index(inplace=True)

        return out

    def _evaluate_filter(self):
        """A new (filtered) table is generated by using a boolen column to select rows."""
        definition = self.definition

        outputs = definition.get('outputs')
        output_table = self.prosto.get_table(outputs[0])

        #
        # Stage 1. Prepare input data
        #
        tables = self.definition.get("tables")
        if not tables:
            log.error("Table filter operation must specify one base table in the 'tables' field.".format())
            return
        tables = self.prosto.get_tables(tables)

        base_table = tables[0]
        base_table_data = base_table.get_data()

        #
        # Stage 2. Find filter column
        #
        columns = self.definition.get("columns")
        if not columns:
            log.error("Filter operation must specify a boolean column from the base table in the 'columns' field.".format())
            return

        filter_column_name = columns[0]
        filter_column = base_table.get_column_data(filter_column_name)

        #
        # Stage 3. Find name of the super link-attribute which will point from this table to the base table
        #
        attributes = output_table.definition.get("attributes", [])
        if len(attributes) != 1:
            log.error("Filter table must declare one attribute for linking to the base table.".format())
            return
        super_attribute = attributes[0]

        #
        # Stage 4. Apply filter
        #
        out = base_table_data[filter_column]

        #
        # Stage 5. Convert base index into a 'super' link attribute
        #
        out = pd.DataFrame(index=out.index)  # We need only index
        out.reset_index(inplace=True, drop=False)  # Convert index into a normal column
        out.rename(columns={'index': super_attribute}, inplace=True)  # Every new table has its own 0-based index

        return out

    def _evaluate_project(self):
        """Find unique combinations of the projected columns and store them as attributes in the populated table."""
        definition = self.definition

        outputs = definition.get('outputs')
        output_table = self.prosto.get_table(outputs[0])

        #
        # Stage 1. Prepare input data
        #
        tables = self.definition.get("tables")
        if not tables:
            log.error("Project operation must specify one input table in the 'tables' field".format())
            return
        tables = self.prosto.get_tables(tables)

        source_table = tables[0]
        source_table_data = source_table.get_data()

        #
        # Stage 2. Find link column
        #
        link_column_name = definition.get('link')
        if not link_column_name:
            log.error("Project operation must specify a link column from the input table in the 'function' field.".format())
            return
        link_column = source_table.get_column(link_column_name)

        #
        # Stage 3. Find parameters of projection: source keys and target attribute names
        #

        # Find source table keys to be projected
        link_column_ops = self.prosto.get_column_operations(source_table.id, link_column.id)
        link_column_op = link_column_ops[0]
        source_keys = link_column_op.get_columns()
        if not all_columns_exist(source_keys, source_table_data):
            log.error("Not all key columns available in the link column definition.".format())
            return

        # Find this (target) table attributes to be created
        attributes = output_table.definition.get("attributes", [])
        if len(attributes) != len(source_keys):
            log.error("Project table must declare one attribute for each input of the corresponding link column.".format())
            return

        # Link column can define its own target keys. We do not use them but want to check the validity because the link evaluation can fail later.
        linked_inputs = link_column_op.definition.get("linked_inputs", [])
        if len(linked_inputs) > 0 and set(attributes) != set(linked_inputs):
            log.warning("Attributes of the project table are different from the 'linked_inputs' of the corresonding link column.".format())

        #
        # Stage 4. Produce all unique combinations of the input columns
        #
        """
        INFO:
        df_new = df.drop_duplicates(subset=['C1', 'C2', 'C3'])  # Drop duplicates
        a_df = df.drop_duplicates(['col1', 'col2'])[['col1', 'col2']]
        df = df.groupby(by=['C1', 'C2', 'C3'], as_index=False).first()  # Using groupby
        np.unique(df[['col1', 'col2']], axis=0)  # Not for object data (error for object types)
        """
        out = source_table.data.get_df().drop_duplicates(subset=source_keys)  # Really do projection

        #
        # Stage 5. Index and renamings
        #
        out = out[source_keys]  # Leave only project columns (de-duplicate will return *all* source columns)

        # Rename to attribute names (de-duplicate will return source table columns)
        rename_dict = dict(zip(source_keys, attributes))  # # source_keys (keys) -> attribute_names (values)
        out.rename(columns=rename_dict, inplace=True)

        out.reset_index(drop=True, inplace=True)  # Every new table has its own 0-based index

        return out


if __name__ == "__main__":
    pass



from typing import Union, Any, List, Set, Dict, Tuple, Optional
import json
from collections import namedtuple

from prosto.utils import *

import logging
log = logging.getLogger('prosto')


Range = namedtuple('Range', 'start end')

class Data:
    """The class represents data physically stored as a pandas data frame."""
    # INFO:
    # d = self.df.iloc[0].name
    # d = self.df.index[0]
    # id = self.df.index.min()
    # id = self.df.first_valid_index()
    # max_id = self.df.index.max()
    # max_id = self.df.last_valid_index()

    data_no = 0

    def __init__(self, table):
        """
        Create a new table object using name and attributes.

        :param name: Name of the data object which should be equal to the table name it represents
        :param definition: Table attributes
        """

        # Assign id
        self.id = None
        if self.id is None:
            self.id = "___data___" + str(Data.data_no)
            Data.data_no += 1

        # Store table it belongs to
        self.table = table

        # Data frame which store the real data for this table (all its attributes and columns)
        attributes = table.definition.get("attributes", [])
        self.df = pd.DataFrame(columns=attributes, index=pd.Int64Index([]))
        self.df.name = table.id
        self.df.index.astype(int)

        # Track changes
        self.removed_range = Range(0, 0)
        self.added_range = Range(0, 0)

    def __repr__(self):
        return '['+self.id+']'

    def get_df(self) -> pd.DataFrame:
        return self.df

    def set_df(self, df) -> None:
        self.df = df
        self.df.name = self.table.id

    def get_series(self, column_name) -> pd.Series:
        return self.df[column_name]

    def all_columns_exist(self, names) -> bool:
        columns = self.df.columns
        for col in names:
            if col not in columns:
                return False
        return True

    #
    # Read column data
    #
    def get_values(self, column_name) -> pd.Series:
        """Read column values"""
        return self.df[column_name]

    def get_full_slice(self, columns) -> pd.DataFrame:
        """Get a slice with all rows (without removed) and specified columns"""

        start_id = self.removed_range.end
        end_id = self.added_range.end

        ret = self.df.loc[start_id:end_id, columns]

        return ret

    def get_added_slice(self, columns) -> pd.DataFrame:
        """Get a slice with added rows and specified columns"""

        start_id = self.added_range.start
        end_id = self.added_range.end

        ret = self.df.loc[start_id:end_id, columns]

        return ret

    #
    # Write column data
    #

    def set_column_values_for_range(self, update, range, default_value) -> int:
        """
        Impose columns from the specified data frame onto this data by overwriting existing cells using index for both columns and rows.
        Set new values for the specified range by overwriting existing values by those from the update frame or (if they are absent) by default value (which can be NaN).
        It is guaranteed that values in the specified range will be changed, that is, old values from the specified range will be removed.
        New values are taken either from the update frame or default value.
        If a column is absent in the target then, it will be added.
        If a row is absent in the target then, it will NOT be added.
        If a row absent in the source, then it will not be updated.
        """

        # TODO: Shorten update frame to the specified range if necessary (or at least check that it is inside the range and warn if not)

        #
        # Update columns by ensuring that new columns exist
        #
        for col in update.columns.to_list():
            if col not in self.df.columns.to_list():
                self.df[col] = default_value

        #
        # Update values
        #

        if range is None:
            range = self.id_range()  # Full range

        # Approach 1:
        # 1) Assign default value to the data in the specified full range (essentially do reset)
        self.df.loc[range.start:range.end, update.columns.to_list()] = default_value
        # 2) Impose values from the update frame on the data. Missing values will not be changed and hence will be equal to default value.
        self.df.update(update, overwrite=True)
        # INFO: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.update.html
        # - if source is NA, then it is ignored (independent of the target)
        # - if target is NA: if overwrite=True then overwrite, elif overwrite=False then leave NA

        # Approach 2:
        # 1) Extend the update index to the specified (full) range by setting added rows to the default value
        # 2) Assign this extended update frame to the data using its index (which has full necessary range)

        #
        # Theoretically, since we change the values, we need to mark this range as changed/dirty
        # Yet, tracking value updates is currently not supported
        #

        return range.end - range.start

    #
    # Add rows
    #

    def add(self) -> int:
        """Add one row and return its id. All attributes and columns will have empty values (None)."""
        empty_value = None
        first_id = self._get_next_id()

        # Approach 1. Specify new index value explicitly
        self.df.loc[first_id, :] = empty_value

        # Approach 2. Series name will be used as new index value
        #self.df = self.df.append(pd.Series(name=first_id, data=[empty_value]), sort=False)

        # Track changes
        self.extend_added(1)

        return first_id

    def add(self, count) -> int:
        """Add several rows."""
        empty_value = None
        first_id = self._get_next_id()

        # Empty data frame with new (added) row ids in the index
        new_ids = range(first_id, first_id + count)
        table = pd.DataFrame(index=new_ids)

        # Approach 1
        self.df = self.df.append(table, sort=False)

        # Approach 2
        #self.df = pd.concat([self.df, table])

        # Track changes
        self.extend_added(count)

        return first_id

    def add(self, record) -> int:
        """Add one new row with the specified attribute values passed as a dictionary or series"""
        empty_value = None
        first_id = self._get_next_id()

        # Approach 1. Specify new index value explicitly
        self.df.loc[first_id, :] = record  # We can assign both Series and dict types

        # Approach 2. Series name will be used as new index value
        #if isinstance(record, dict):
        #    record = pd.Series(record, name=first_id)
        #self.df = self.df.append(record, sort=False)

        # Track changes
        self.extend_added(1)

        return first_id

    def add(self, table) -> int:
        """Add multiple new rows with the specified attribute values passed as a structured which is a dataframe or can be used to instantiate a data frame."""
        empty_value = None
        first_id = self._get_next_id()

        # Data frame with new (added) row ids in the index and data to be appended
        count = len(table)
        new_ids = range(first_id, first_id + count)
        table = pd.DataFrame(table, index=new_ids)  # Even if it is already a data frame, we want to explicitly set its index

        # Approach 1
        self.df = self.df.append(table, sort=False)

        # Approach 2
        #self.df = pd.concat([self.df, table])

        # Track changes
        self.extend_added(count)

        return first_id

    #
    # Physically delete records and manage allocated space
    #

    def gc(self) -> None:
        """Physically delete all records which are not used, that is, their removal was already propagated."""
        if self.df.empty:
            return
        start = self.df.index[0]
        end = self.removed_range.start
        to_delete = range(start, end)
        self.df.drop(to_delete, inplace=True)
        #self.df = self.df.iloc[len(to_delete):]

    def reset(self) -> None:
        """Physically remove all records and start from new empty table with no tracking."""

        #self.df.drop(self.df.index, inplace=True)
        #self.df = pd.DataFrame(columns=self.df.columns)
        self.df = self.df.iloc[0:0]

        self.df.name = self.table.id  # name is not copied for some reason

        # Track changes
        self.added_range = Range(0, 0)
        self.removed_range = Range(0, 0)

    #
    # Track changes
    #

    def id_range(self) -> Range:
        return Range(self.removed_range.end, self.added_range.end)

    def length(self) -> int:
        return self.added_range.end - self.removed_range.end  # All non-removed records

    def added_length(self) -> int:
        return self.added_range.end - self.added_range.start

    def removed_length(self) -> int:
        return self.removed_range.end - self.removed_range.start

    def extend_added(self, count=1) -> None:
        self.added_range = Range(self.added_range.start, self.added_range.end + count)

    def extend_removed(self, count=1) -> None:
        self.removed_range = Range(self.removed_range.start, self.removed_range.end + count)

    def shrink_added(self) -> int:
        added = self.added_range.end - self.added_range.start
        self.added_range = Range(self.added_range.end, self.added_range.end)
        return added

    def shrink_removed(self) -> int:
        removed = self.removed_range.end - self.removed_range.start
        self.removed_range = Range(self.removed_range.end, self.removed_range.end)
        return removed

    def clear_change_status(self) -> None:
        added = self.shrink_added()
        removed = self.shrink_removed()

    #
    # Remove rows (mark for removal)
    #

    def remove(self) -> None:
        """Mark one oldest record as removed."""

        if self.length() > 0:
            # Extend removed range by 1
            self.removed_range.end = Range(self.removed_range.start, self.removed_range.end + 1)

    def remove(self, count) -> Range:
        """Mark the specified number of oldest records as removed."""

        to_remove = min(count, self.length())
        if to_remove > 0:
            # Extend removed range by count
            self.removed_range = Range(self.removed_range.start, self.removed_range.end + to_remove)

        return Range(self.removed_range.end - to_remove, self.removed_range.end)

    def remove_all(self) -> None:
        """Mark all records as removed."""

        # Track changes
        if self.length() > 0:
            # Extend removed range till the end of added range (added records are also removed)
            self.removed_range = Range(self.removed_range.start, self.added_range.end)

    #
    # Convenience methods
    #

    def _get_next_id(self)  -> int:
        return self.added_range.end

    def _get_start_offset(self) -> int:
        """Physically existing records"""
        return 0

    def _get_end_offset(self) -> int:
        """Physically existing records"""
        return len(self.df)


if __name__ == "__main__":
    pass
